var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
define(["require", "exports", "../../../release/go", "./WallBuildingTool", "./WallReshapingTool"], function (require, exports, go, WallBuildingTool, WallReshapingTool) {
    "use strict";
    var Floorplan = (function (_super) {
        __extends(Floorplan, _super);
        function Floorplan(div) {
            var _this = _super.call(this, div) || this;
            _this.convertPixelsToUnits = function (num) {
                var units = this.model.modelData.units;
                var factor = this.model.modelData.unitsConversionFactor;
                if (units === 'meters')
                    return (num / 100) * factor;
                if (units === 'feet')
                    return (num / 30.48) * factor;
                if (units === 'inches')
                    return (num / 2.54) * factor;
                return num * factor;
            };
            _this.convertUnitsToPixels = function (num) {
                var units = this.model.modelData.units;
                var factor = this.model.modelData.unitsConversionFactor;
                if (units === 'meters')
                    return (num * 100) / factor;
                if (units === 'feet')
                    return (num * 30.48) / factor;
                if (units === 'inches')
                    return (num * 2.54) / factor;
                return num / factor;
            };
            _this.updateWall = function (wall) {
                if (wall.data.startpoint && wall.data.endpoint) {
                    var shape = wall.findObject("SHAPE");
                    var geo = new go.Geometry(go.Geometry.Line);
                    var sPt = wall.data.startpoint;
                    var ePt = wall.data.endpoint;
                    var mPt = new go.Point((sPt.x + ePt.x) / 2, (sPt.y + ePt.y) / 2);
                    geo.startX = 0;
                    geo.startY = 0;
                    geo.endX = Math.sqrt(sPt.distanceSquaredPoint(ePt));
                    geo.endY = 0;
                    shape.geometry = geo;
                    wall.location = mPt;
                    var angle = sPt.directionPoint(ePt);
                    wall.rotateObject.angle = angle;
                    this.updateWallDimensions();
                }
            };
            _this.getAdjustedPoint = function (point, wall, angle, wallOffset) {
                var oldPoint = point.copy();
                point.offset(0, -(wall.data.thickness * .5) - wallOffset);
                point.offset(-oldPoint.x, -oldPoint.y).rotate(angle).offset(oldPoint.x, oldPoint.y);
                return point;
            };
            _this.buildDimensionLink = function (wall, index, point1, point2, angle, wallOffset, soloWallFlag, floorplan) {
                point1 = floorplan.getAdjustedPoint(point1, wall, angle, wallOffset);
                point2 = floorplan.getAdjustedPoint(point2, wall, angle, wallOffset);
                var data1 = { key: wall.data.key + "PointNode" + index, category: "PointNode", loc: go.Point.stringify(point1) };
                var data2 = { key: wall.data.key + "PointNode" + (index + 1), category: "PointNode", loc: go.Point.stringify(point2) };
                var data3 = { key: wall.data.key + "DimensionLink", category: 'DimensionLink', from: data1.key, to: data2.key, stroke: 'gray', angle: angle, wall: wall.data.key, soloWallFlag: soloWallFlag };
                var pointNode1 = makePointNode();
                var pointNode2 = makePointNode();
                var link = makeDimensionLink();
                floorplan.pointNodes.add(pointNode1);
                floorplan.pointNodes.add(pointNode2);
                floorplan.dimensionLinks.add(link);
                floorplan.add(pointNode1);
                floorplan.add(pointNode2);
                floorplan.add(link);
                pointNode1.data = data1;
                pointNode2.data = data2;
                link.data = data3;
                link.fromNode = pointNode1;
                link.toNode = pointNode2;
            };
            _this.updateWallDimensions = function () {
                var floorplan = this;
                floorplan.skipsUndoManager = true;
                floorplan.startTransaction("update wall dimensions");
                if (!floorplan.model.modelData.preferences.showWallLengths) {
                    floorplan.pointNodes.iterator.each(function (node) { floorplan.remove(node); });
                    floorplan.dimensionLinks.iterator.each(function (link) { floorplan.remove(link); });
                    floorplan.pointNodes.clear();
                    floorplan.dimensionLinks.clear();
                    floorplan.commitTransaction("update wall dimensions");
                    floorplan.skipsUndoManager = false;
                    return;
                }
                floorplan.dimensionLinks.iterator.each(function (link) { link.visible = true; });
                var selection = floorplan.selection;
                var walls = new go.Set();
                selection.iterator.each(function (part) {
                    if ((part.category === 'WindowNode' || part.category === 'DoorNode') && part.containingGroup !== null)
                        walls.add(part.containingGroup);
                    if (part.category === 'WallGroup' && part.data && part.data.startpoint && part.data.endpoint) {
                        var wall_1 = part;
                        var soloWallLink_1 = null;
                        floorplan.dimensionLinks.iterator.each(function (link) { if (link.data.soloWallFlag && link.data.wall === wall_1.data.key)
                            soloWallLink_1 = link; });
                        if (soloWallLink_1 !== null) {
                            var linkPoint1_1 = null;
                            var linkPoint2_1 = null;
                            floorplan.pointNodes.iterator.each(function (node) {
                                if (node.data.key === wall_1.data.key + "PointNode1")
                                    linkPoint1_1 = node;
                                if (node.data.key === wall_1.data.key + "PointNode2")
                                    linkPoint2_1 = node;
                            });
                            var startpoint = wall_1.data.startpoint;
                            var endpoint = wall_1.data.endpoint;
                            var firstWallPt = ((startpoint.x + startpoint.y) <= (endpoint.x + endpoint.y)) ? startpoint : endpoint;
                            var lastWallPt = ((startpoint.x + startpoint.y) > (endpoint.x + endpoint.y)) ? startpoint : endpoint;
                            var newLoc1 = floorplan.getAdjustedPoint(firstWallPt.copy(), wall_1, wall_1.rotateObject.angle, 10);
                            var newLoc2 = floorplan.getAdjustedPoint(lastWallPt.copy(), wall_1, wall_1.rotateObject.angle, 10);
                            linkPoint1_1.data.loc = go.Point.stringify(newLoc1);
                            linkPoint2_1.data.loc = go.Point.stringify(newLoc2);
                            soloWallLink_1.data.angle = wall_1.rotateObject.angle;
                            linkPoint1_1.updateTargetBindings();
                            linkPoint2_1.updateTargetBindings();
                            soloWallLink_1.updateTargetBindings();
                        }
                        else {
                            var startpoint = wall_1.data.startpoint;
                            var endpoint = wall_1.data.endpoint;
                            var firstWallPt = ((startpoint.x + startpoint.y) <= (endpoint.x + endpoint.y)) ? startpoint : endpoint;
                            var lastWallPt = ((startpoint.x + startpoint.y) > (endpoint.x + endpoint.y)) ? startpoint : endpoint;
                            floorplan.buildDimensionLink(wall_1, 1, firstWallPt.copy(), lastWallPt.copy(), wall_1.rotateObject.angle, 10, true, floorplan);
                        }
                    }
                });
                walls.iterator.each(function (wall) {
                    var startpoint = wall.data.startpoint;
                    var endpoint = wall.data.endpoint;
                    var firstWallPt = ((startpoint.x + startpoint.y) <= (endpoint.x + endpoint.y)) ? startpoint : endpoint;
                    var lastWallPt = ((startpoint.x + startpoint.y) > (endpoint.x + endpoint.y)) ? startpoint : endpoint;
                    var wallPartEndpoints = [];
                    wall.memberParts.iterator.each(function (wallPart) {
                        if (wallPart.isSelected) {
                            var endpoints = getWallPartEndpoints(wallPart);
                            wallPartEndpoints.push(endpoints[0]);
                            wallPartEndpoints.push(endpoints[1]);
                        }
                    });
                    wallPartEndpoints.sort(function (a, b) {
                        if ((a.x + a.y) > (b.x + b.y))
                            return 1;
                        if ((a.x + a.y) < (b.x + b.y))
                            return -1;
                        else
                            return 0;
                    });
                    wallPartEndpoints.unshift(firstWallPt);
                    wallPartEndpoints.push(lastWallPt);
                    var angle = wall.rotateObject.angle;
                    var k = 1;
                    var _loop_1 = function (j) {
                        var linkPoint1 = null;
                        var linkPoint2 = null;
                        floorplan.pointNodes.iterator.each(function (node) {
                            if (node.data.key === wall.data.key + "PointNode" + k)
                                linkPoint1 = node;
                            if (node.data.key === wall.data.key + "PointNode" + (k + 1))
                                linkPoint2 = node;
                        });
                        if (linkPoint1 !== null) {
                            var newLoc1 = floorplan.getAdjustedPoint(wallPartEndpoints[j].copy(), wall, angle, 5);
                            var newLoc2 = floorplan.getAdjustedPoint(wallPartEndpoints[j + 1].copy(), wall, angle, 5);
                            linkPoint1.data.loc = go.Point.stringify(newLoc1);
                            linkPoint2.data.loc = go.Point.stringify(newLoc2);
                            linkPoint1.updateTargetBindings();
                            linkPoint2.updateTargetBindings();
                        }
                        else
                            floorplan.buildDimensionLink(wall, k, wallPartEndpoints[j].copy(), wallPartEndpoints[j + 1].copy(), angle, 5, false, floorplan);
                        k += 2;
                    };
                    for (var j = 0; j < wallPartEndpoints.length - 1; j++) {
                        _loop_1(j);
                    }
                    var totalWallDimensionLink = null;
                    floorplan.dimensionLinks.iterator.each(function (link) {
                        if ((link.fromNode.data.key === wall.data.key + "PointNode" + k) &&
                            (link.toNode.data.key === wall.data.key + "PointNode" + (k + 1)))
                            totalWallDimensionLink = link;
                    });
                    if (totalWallDimensionLink !== null) {
                        var linkPoint1_2 = null;
                        var linkPoint2_2 = null;
                        floorplan.pointNodes.iterator.each(function (node) {
                            if (node.data.key === wall.data.key + "PointNode" + k)
                                linkPoint1_2 = node;
                            if (node.data.key === wall.data.key + "PointNode" + (k + 1))
                                linkPoint2_2 = node;
                        });
                        var newLoc1 = floorplan.getAdjustedPoint(wallPartEndpoints[0].copy(), wall, angle, 25);
                        var newLoc2 = floorplan.getAdjustedPoint(wallPartEndpoints[wallPartEndpoints.length - 1].copy(), wall, angle, 25);
                        linkPoint1_2.data.loc = go.Point.stringify(newLoc1);
                        linkPoint2_2.data.loc = go.Point.stringify(newLoc2);
                        linkPoint1_2.updateTargetBindings();
                        linkPoint2_2.updateTargetBindings();
                    }
                    else
                        floorplan.buildDimensionLink(wall, k, wallPartEndpoints[0].copy(), wallPartEndpoints[wallPartEndpoints.length - 1].copy(), angle, 25, false, floorplan);
                });
                floorplan.dimensionLinks.iterator.each(function (link) {
                    var canStay = false;
                    floorplan.pointNodes.iterator.each(function (node) {
                        if (node.data.key == link.data.to)
                            canStay = true;
                    });
                    if (!canStay)
                        floorplan.remove(link);
                    else {
                        var length_1 = Math.sqrt(link.toNode.location.distanceSquaredPoint(link.fromNode.location));
                        if (length_1 < 1 && !link.data.soloWallFlag)
                            link.visible = false;
                    }
                });
                floorplan.commitTransaction("update wall dimensions");
                floorplan.skipsUndoManager = false;
            };
            _this.getWallsIntersection = function (wall1, wall2) {
                if (wall1 === null || wall2 === null)
                    return null;
                var a1 = wall1.data.endpoint.y - wall1.data.startpoint.y;
                var b1 = wall1.data.startpoint.x - wall1.data.endpoint.x;
                var c1 = (a1 * wall1.data.startpoint.x) + (b1 * wall1.data.startpoint.y);
                var a2 = wall2.data.endpoint.y - wall2.data.startpoint.y;
                var b2 = wall2.data.startpoint.x - wall2.data.endpoint.x;
                var c2 = (a2 * wall2.data.startpoint.x) + (b2 * wall2.data.startpoint.y);
                var det = a1 * b2 - a2 * b1;
                var x = null;
                var y = null;
                if (det === 0) {
                    if (wall1.data.startpoint.equals(wall2.data.startpoint) || wall1.data.startpoint.equals(wall2.data.endpoint))
                        return wall1.data.startpoint;
                    if (wall1.data.endpoint.equals(wall2.data.startpoint) || wall1.data.endpoint.equals(wall2.data.endpoint))
                        return wall1.data.endpoint;
                    return null;
                }
                else {
                    x = (b2 * c1 - b1 * c2) / det;
                    y = (a1 * c2 - a2 * c1) / det;
                }
                var inWall1 = ((Math.min(wall1.data.startpoint.x, wall1.data.endpoint.x) <= x) && (Math.max(wall1.data.startpoint.x, wall1.data.endpoint.x) >= x)
                    && (Math.min(wall1.data.startpoint.y, wall1.data.endpoint.y) <= y) && (Math.max(wall1.data.startpoint.y, wall1.data.endpoint.y) >= y));
                var inWall2 = ((Math.min(wall2.data.startpoint.x, wall2.data.endpoint.x) <= x) && (Math.max(wall2.data.startpoint.x, wall2.data.endpoint.x) >= x)
                    && (Math.min(wall2.data.startpoint.y, wall2.data.endpoint.y) <= y) && (Math.max(wall2.data.startpoint.y, wall2.data.endpoint.y) >= y));
                if (inWall1 && inWall2)
                    return new go.Point(x, y);
                else
                    return null;
            };
            _this.updateWallAngles = function () {
                var floorplan = this;
                floorplan.skipsUndoManager = true;
                floorplan.startTransaction("display angles");
                if (floorplan.model.modelData.preferences.showWallAngles) {
                    floorplan.angleNodes.iterator.each(function (node) { node.visible = true; });
                    var selectedWalls_1 = [];
                    floorplan.selection.iterator.each(function (part) {
                        if (part.category === "WallGroup") {
                            var w = part;
                            selectedWalls_1.push(w);
                        }
                    });
                    var _loop_2 = function (i) {
                        var seen = new go.Set();
                        var wall = selectedWalls_1[i];
                        var possibleWalls = floorplan.findNodesByExample({ category: "WallGroup" });
                        possibleWalls.iterator.each(function (otherWall) {
                            if (otherWall.data === null || wall.data === null || seen.contains(otherWall.data.key))
                                return;
                            if ((otherWall.data.key !== wall.data.key) && (floorplan.getWallsIntersection(wall, otherWall) !== null) && (!seen.contains(otherWall.data.key))) {
                                seen.add(otherWall.data.key);
                                var intersectionPoint_1 = floorplan.getWallsIntersection(wall, otherWall);
                                var wallsInvolved = floorplan.findObjectsNear(intersectionPoint_1, 1, function (x) { if (x.part !== null)
                                    return x.part; }, function (p) {
                                    if (!(p instanceof go.Group))
                                        return false;
                                    else
                                        return p.category === "WallGroup";
                                }, false);
                                var endpoints_1 = [];
                                wallsInvolved.iterator.each(function (w) {
                                    var tolerance = (floorplan.model.modelData.gridSize >= 10) ? floorplan.model.modelData.gridSize : 10;
                                    if (Math.sqrt(w.data.startpoint.distanceSquaredPoint(intersectionPoint_1)) > tolerance)
                                        endpoints_1.push({ point: w.data.startpoint, wall: w.data.key });
                                    if (Math.sqrt(w.data.endpoint.distanceSquaredPoint(intersectionPoint_1)) > tolerance)
                                        endpoints_1.push({ point: w.data.endpoint, wall: w.data.key });
                                });
                                var maxRadius = 30;
                                for (var i_1 = 0; i_1 < endpoints_1.length; i_1++) {
                                    var distance = Math.sqrt(endpoints_1[i_1].point.distanceSquaredPoint(intersectionPoint_1));
                                    if (distance < maxRadius)
                                        maxRadius = distance;
                                }
                                endpoints_1.sort(function (a, b) {
                                    a = a.point;
                                    b = b.point;
                                    if (a.x - intersectionPoint_1.x >= 0 && b.x - intersectionPoint_1.x < 0)
                                        return 1;
                                    if (a.x - intersectionPoint_1.x < 0 && b.x - intersectionPoint_1.x >= 0)
                                        return -1;
                                    if (a.x - intersectionPoint_1.x == 0 && b.x - intersectionPoint_1.x == 0) {
                                        if (a.y - intersectionPoint_1.y >= 0 || b.y - intersectionPoint_1.y >= 0)
                                            return a.y > b.y ? 1 : -1;
                                        return b.y > a.y ? 1 : -1;
                                    }
                                    var det = (a.x - intersectionPoint_1.x) * (b.y - intersectionPoint_1.y) - (b.x - intersectionPoint_1.x) * (a.y - intersectionPoint_1.y);
                                    if (det < 0)
                                        return 1;
                                    if (det > 0)
                                        return -1;
                                    var d1 = (a.x - intersectionPoint_1.x) * (a.x - intersectionPoint_1.x) + (a.y - intersectionPoint_1.y) * (a.y - intersectionPoint_1.y);
                                    var d2 = (b.x - intersectionPoint_1.x) * (b.x - intersectionPoint_1.x) + (b.y - intersectionPoint_1.y) * (b.y - intersectionPoint_1.y);
                                    return d1 > d2 ? 1 : -1;
                                });
                                var _loop_3 = function (i_2) {
                                    var p1 = endpoints_1[i_2];
                                    var p2 = void 0;
                                    if (endpoints_1[i_2 + 1] != null) {
                                        p2 = endpoints_1[i_2 + 1];
                                    }
                                    else {
                                        p2 = endpoints_1[0];
                                    }
                                    var a1 = intersectionPoint_1.directionPoint(p1.point);
                                    var a2 = intersectionPoint_1.directionPoint(p2.point);
                                    var sweep = Math.abs(a2 - a1 + 360) % 360;
                                    var angle = a1;
                                    var keyArray = [];
                                    wallsInvolved.iterator.each(function (wall) { keyArray.push(wall); });
                                    keyArray.sort(function (a, b) {
                                        var aIndex = a.data.key.match(/\d+/g);
                                        var bIndex = b.data.key.match(/\d+/g);
                                        if (isNaN(aIndex))
                                            return 1;
                                        if (isNaN(bIndex))
                                            return -1;
                                        else
                                            return aIndex > bIndex ? 1 : -1;
                                    });
                                    var key = "";
                                    for (var j = 0; j < keyArray.length; j++)
                                        key += keyArray[j].data.key;
                                    key += "angle" + i_2;
                                    var angleNode = null;
                                    floorplan.angleNodes.iterator.each(function (aNode) { if (aNode.data.key === key)
                                        angleNode = aNode; });
                                    if (angleNode !== null) {
                                        angleNode.data.angle = angle;
                                        angleNode.data.sweep = sweep;
                                        angleNode.data.loc = go.Point.stringify(intersectionPoint_1);
                                        angleNode.data.maxRadius = maxRadius;
                                        angleNode.updateTargetBindings();
                                    }
                                    else {
                                        var data = { key: key, category: "AngleNode", loc: go.Point.stringify(intersectionPoint_1), stroke: "dodgerblue", angle: angle, sweep: sweep, maxRadius: maxRadius };
                                        var newAngleNode = makeAngleNode();
                                        newAngleNode.data = data;
                                        floorplan.add(newAngleNode);
                                        newAngleNode.updateTargetBindings();
                                        floorplan.angleNodes.add(newAngleNode);
                                    }
                                };
                                for (var i_2 = 0; i_2 < endpoints_1.length; i_2++) {
                                    _loop_3(i_2);
                                }
                            }
                        });
                    };
                    for (var i = 0; i < selectedWalls_1.length; i++) {
                        _loop_2(i);
                    }
                    var garbage_1 = [];
                    floorplan.angleNodes.iterator.each(function (node) {
                        var keyNums = node.data.key.match(/\d+/g);
                        var numWalls = (node.data.key.match(/wall/g) || []).length;
                        var wallsInvolved = [];
                        for (var i = 0; i < keyNums.length - 1; i++)
                            wallsInvolved.push("wall" + keyNums[i]);
                        if (numWalls !== keyNums.length - 1)
                            wallsInvolved.push("wall");
                        for (var i = 0; i < wallsInvolved.length - 1; i++) {
                            var wall1 = floorplan.findPartForKey(wallsInvolved[i]);
                            var wall2 = floorplan.findPartForKey(wallsInvolved[i + 1]);
                            var intersectionPoint = floorplan.getWallsIntersection(wall1, wall2);
                            if (intersectionPoint === null)
                                garbage_1.push(node);
                        }
                        var possibleAngleNodes = new go.Set();
                        var allWalls = node.data.key.slice(0, node.data.key.indexOf("angle"));
                        floorplan.angleNodes.iterator.each(function (other) { if (other.data.key.indexOf(allWalls) !== -1)
                            possibleAngleNodes.add(other); });
                        possibleAngleNodes.iterator.each(function (pNode) {
                            if (pNode.data.loc !== node.data.loc) {
                                garbage_1.push(pNode);
                            }
                        });
                        if (node.data.sweep === 0)
                            garbage_1.push(node);
                    });
                    for (var i = 0; i < garbage_1.length; i++) {
                        floorplan.remove(garbage_1[i]);
                        floorplan.angleNodes.remove(garbage_1[i]);
                    }
                }
                if (floorplan.model.modelData.preferences.showOnlySmallWallAngles) {
                    floorplan.angleNodes.iterator.each(function (node) { if (node.data.sweep >= 180)
                        node.visible = false; });
                }
                if (!floorplan.model.modelData.preferences.showWallAngles) {
                    floorplan.angleNodes.iterator.each(function (node) { node.visible = false; });
                }
                floorplan.commitTransaction("display angles");
                floorplan.skipsUndoManager = false;
            };
            _this._palettes = [];
            _this._pointNodes = new go.Set();
            _this._dimensionLinks = new go.Set();
            _this._angleNodes = new go.Set();
            var $ = go.GraphObject.make;
            _this.allowLink = false;
            _this.undoManager.isEnabled = true;
            _this.layout.isOngoing = false;
            _this.model = $(go.GraphLinksModel, {
                modelData: {
                    "units": "centimeters",
                    "unitsAbbreviation": "cm",
                    "gridSize": 10,
                    "wallThickness": 5,
                    "preferences": {
                        showWallGuidelines: true,
                        showWallLengths: true,
                        showWallAngles: true,
                        showOnlySmallWallAngles: true,
                        showGrid: true,
                        gridSnap: true
                    }
                }
            });
            _this.grid = $(go.Panel, "Grid", { gridCellSize: new go.Size(_this.model.modelData.gridSize, _this.model.modelData.gridSize), visible: true }, $(go.Shape, "LineH", { stroke: "lightgray" }), $(go.Shape, "LineV", { stroke: "lightgray" }));
            _this.contextMenu = makeContextMenu();
            _this.commandHandler.canGroupSelection = function () { return true; };
            _this.commandHandler.canUngroupSelection = function () { return true; };
            _this.commandHandler.archetypeGroupData = { isGroup: true };
            _this.addDiagramListener("SelectionCopied", function (e) {
                var fp = e.diagram;
                fp.selection.iterator.each(function (part) {
                    if (part.category == "WallGroup") {
                        var w = part;
                        fp.updateWall(w);
                    }
                });
            });
            _this.addDiagramListener("ExternalObjectsDropped", function (e) {
                var garbage = [];
                var fp = e.diagram;
                fp.selection.iterator.each(function (node) {
                    if (node.category === "PaletteWallNode") {
                        var paletteWallNode = node;
                        var endpoints = getWallPartEndpoints(paletteWallNode);
                        var data = { key: "wall", category: "WallGroup", caption: "Wall", startpoint: endpoints[0], endpoint: endpoints[1], thickness: parseFloat(e.diagram.model.modelData.wallThickness), isGroup: true, notes: "" };
                        e.diagram.model.addNodeData(data);
                        var wall = e.diagram.findPartForKey(data.key);
                        fp.updateWall(wall);
                        garbage.push(paletteWallNode);
                    }
                });
                for (var i in garbage) {
                    e.diagram.remove(garbage[i]);
                }
            });
            _this.addDiagramListener("ClipboardPasted", function (e) {
                var fp = e.diagram;
                e.diagram.selection.iterator.each(function (node) {
                    if (node.category === "WallGroup") {
                        var w = node;
                        fp.updateWall(w);
                    }
                });
            });
            _this.nodeTemplateMap.add("", makeDefaultNode());
            _this.nodeTemplateMap.add("MultiPurposeNode", makeMultiPurposeNode());
            _this.nodeTemplateMap.add("WindowNode", makeWindowNode());
            _this.nodeTemplateMap.add("PaletteWallNode", makePaletteWallNode());
            _this.nodeTemplateMap.add("DoorNode", makeDoorNode());
            _this.groupTemplateMap.add("", makeDefaultGroup());
            _this.groupTemplateMap.add("WallGroup", makeWallGroup());
            var wallBuildingTool = new WallBuildingTool();
            _this.toolManager.mouseDownTools.insertAt(0, wallBuildingTool);
            var wallReshapingTool = new WallReshapingTool();
            _this.toolManager.mouseDownTools.insertAt(3, wallReshapingTool);
            wallBuildingTool.isEnabled = false;
            _this.toolManager.draggingTool.doMouseUp = function () {
                go.DraggingTool.prototype.doMouseUp.call(this);
                this.diagram.updateWallAngles();
                this.isGridSnapEnabled = this.diagram.model.modelData.preferences.gridSnap;
            };
            _this.toolManager.draggingTool.doMouseMove = function () {
                if (this.diagram.lastInput.shift) {
                    this.isGridSnapEnabled = false;
                }
                else
                    this.isGridSnapEnabled = this.diagram.model.modelData.preferences.gridSnap;
                go.DraggingTool.prototype.doMouseMove.call(this);
            };
            _this.toolManager.resizingTool.doMouseMove = function () {
                var floorplan = this.diagram;
                var node = this.adornedObject;
                this.diagram.updateWallDimensions();
                go.ResizingTool.prototype.doMouseMove.call(this);
            };
            _this.toolManager.resizingTool.computeMaxSize = function () {
                var tool = this;
                var obj = tool.adornedObject.part;
                var wall = this.diagram.findPartForKey(obj.data.group);
                if ((obj.category === 'DoorNode' || obj.category === 'WindowNode') && wall !== null) {
                    var stationaryPt;
                    var movingPt;
                    var resizeAdornment = null;
                    obj.adornments.iterator.each(function (adorn) { if (adorn.name === "WallPartResizeAdornment")
                        resizeAdornment = adorn; });
                    resizeAdornment.elements.iterator.each(function (el) {
                        if (el instanceof go.Shape && el.alignment === tool.handle.alignment)
                            movingPt = el.getDocumentPoint(go.Spot.Center);
                        if (el instanceof go.Shape && el.alignment !== tool.handle.alignment)
                            stationaryPt = el.getDocumentPoint(go.Spot.Center);
                    });
                    var constrainingPt;
                    var closestDist = Number.MAX_VALUE;
                    wall.memberParts.iterator.each(function (part) {
                        if (part.data.key !== obj.data.key) {
                            var endpoints = getWallPartEndpoints(part);
                            for (var i = 0; i < endpoints.length; i++) {
                                var point = endpoints[i];
                                var distanceToMovingPt = Math.sqrt(point.distanceSquaredPoint(movingPt));
                                if (distanceToMovingPt < closestDist) {
                                    var distanceToStationaryPt = Math.sqrt(point.distanceSquaredPoint(stationaryPt));
                                    if (distanceToStationaryPt > distanceToMovingPt) {
                                        closestDist = distanceToMovingPt;
                                        constrainingPt = point;
                                    }
                                }
                            }
                        }
                    });
                    if (constrainingPt === undefined || constrainingPt === null) {
                        if (wall.data.startpoint.distanceSquaredPoint(movingPt) > wall.data.startpoint.distanceSquaredPoint(stationaryPt))
                            constrainingPt = wall.data.endpoint;
                        else
                            constrainingPt = wall.data.startpoint;
                    }
                    var maxLength = Math.sqrt(stationaryPt.distanceSquaredPoint(constrainingPt));
                    return new go.Size(maxLength, wall.data.thickness);
                }
                return go.ResizingTool.prototype.computeMaxSize.call(tool);
            };
            _this.toolManager.draggingTool.isGridSnapEnabled = true;
            return _this;
        }
        Object.defineProperty(Floorplan.prototype, "palettes", {
            get: function () { return this._palettes; },
            set: function (value) { this._palettes = value; },
            enumerable: true,
            configurable: true
        });
        Object.defineProperty(Floorplan.prototype, "pointNodes", {
            get: function () { return this._pointNodes; },
            set: function (value) { this._pointNodes = value; },
            enumerable: true,
            configurable: true
        });
        Object.defineProperty(Floorplan.prototype, "dimensionLinks", {
            get: function () { return this._dimensionLinks; },
            set: function (value) { this._dimensionLinks = value; },
            enumerable: true,
            configurable: true
        });
        Object.defineProperty(Floorplan.prototype, "angleNodes", {
            get: function () { return this._angleNodes; },
            set: function (value) { this._angleNodes = value; },
            enumerable: true,
            configurable: true
        });
        return Floorplan;
    }(go.Diagram));
    function makeSelectionGroup(floorplan) {
        floorplan.startTransaction("group selection");
        var sel = floorplan.selection;
        var nodes = [];
        sel.iterator.each(function (n) {
            if (n instanceof go.Group)
                n.memberParts.iterator.each(function (part) { nodes.push(part); });
            else
                nodes.push(n);
        });
        for (var i = 0; i < nodes.length; i++)
            nodes[i].isSelected = true;
        ungroupSelection(floorplan);
        floorplan.commandHandler.groupSelection();
        var group = floorplan.selection.first();
        floorplan.model.setDataProperty(group.data, "caption", "Group");
        floorplan.model.setDataProperty(group.data, "notes", "");
        clearEmptyGroups(floorplan);
        floorplan.clearSelection();
        floorplan.select(group);
        floorplan.commitTransaction("group selection");
    }
    function ungroupSelection(floorplan) {
        floorplan.startTransaction('ungroup selection');
        function ungroupNode(node) {
            var group = node.containingGroup;
            node.containingGroup = null;
            if (group != null) {
                if (group.memberParts.count === 0)
                    floorplan.remove(group);
                else if (group.memberParts.count === 1)
                    group.memberParts.first().containingGroup = null;
            }
        }
        var sel = floorplan.selection;
        var groups = [];
        sel.iterator.each(function (n) {
            if (!(n instanceof go.Group))
                ungroupNode(n);
            else
                groups.push(n);
        });
        var nodes = [];
        for (var i = 0; i < groups.length; i++)
            groups[i].memberParts.iterator.each(function (n) { nodes.push(n); });
        for (var i = 0; i < nodes.length; i++)
            ungroupNode(nodes[i]);
        clearEmptyGroups(floorplan);
        floorplan.commitTransaction('ungroup selection');
    }
    function clearEmptyGroups(floorplan) {
        var nodes = floorplan.nodes;
        var arr = [];
        nodes.iterator.each(function (node) { if (node instanceof go.Group && node.memberParts.count === 0 && node.category !== "WallGroup") {
            arr.push(node);
        } });
        for (var i = 0; i < arr.length; i++) {
            floorplan.remove(arr[i]);
        }
    }
    function makeGroupToolTip() {
        var $ = go.GraphObject.make;
        return $(go.Adornment, "Auto", $(go.Shape, { fill: "#FFFFCC" }), $(go.TextBlock, { margin: 4 }, new go.Binding("text", "", function (text, obj) {
            var data = obj.part.adornedObject.data;
            var name = (obj.part.adornedObject.category === "MultiPurposeNode") ? data.text : data.caption;
            return "Name: " + name + "\nNotes: " + data.notes + '\nMembers: ' + obj.part.adornedObject.memberParts.count;
        }).ofObject()));
    }
    function makeContextMenu() {
        var $ = go.GraphObject.make;
        return $(go.Adornment, "Vertical", $("ContextMenuButton", $(go.TextBlock, "Make Group"), { click: function (e, obj) { makeSelectionGroup(obj.part.diagram); } }, new go.Binding("visible", "visible", function (v, obj) {
            var floorplan = obj.part.diagram;
            if (floorplan.selection.count <= 1)
                return false;
            var flag = true;
            floorplan.selection.iterator.each(function (node) {
                if (node.category === "WallGroup" || node.category === "WindowNode" || node.category === "DoorNode")
                    flag = false;
            });
            return flag;
        }).ofObject()), $("ContextMenuButton", $(go.TextBlock, "Ungroup"), { click: function (e, obj) { ungroupSelection(obj.part.diagram); } }, new go.Binding("visible", "", function (v, obj) {
            var floorplan = obj.part.diagram;
            if (floorplan !== null) {
                var node = floorplan.selection.first();
                return ((node instanceof go.Node && node.containingGroup != null && node.containingGroup.category != 'WallGroup') ||
                    (node instanceof go.Group && node.category === ''));
            }
            return false;
        }).ofObject()), $("ContextMenuButton", $(go.TextBlock, "Copy"), { click: function (e, obj) { obj.part.diagram.commandHandler.copySelection(); } }, new go.Binding("visible", "", function (v, obj) {
            if (obj.part.diagram !== null) {
                return obj.part.diagram.selection.count > 0;
            }
            return false;
        }).ofObject()), $("ContextMenuButton", $(go.TextBlock, "Cut"), { click: function (e, obj) { obj.part.diagram.commandHandler.cutSelection(); } }, new go.Binding("visible", "", function (v, obj) {
            if (obj.part.diagram !== null) {
                return obj.part.diagram.selection.count > 0;
            }
            return false;
        }).ofObject()), $("ContextMenuButton", $(go.TextBlock, "Delete"), { click: function (e, obj) { obj.part.diagram.commandHandler.deleteSelection(); } }, new go.Binding("visible", "", function (v, obj) {
            if (obj.part.diagram !== null) {
                return obj.part.diagram.selection.count > 0;
            }
            return false;
        }).ofObject()), $("ContextMenuButton", $(go.TextBlock, "Paste"), { click: function (e, obj) { obj.part.diagram.commandHandler.pasteSelection(obj.part.diagram.lastInput.documentPoint); } }), $("ContextMenuButton", $(go.TextBlock, "Show Selection Info"), {
            click: function (e, obj) {
                if (e.diagram.floorplanUI) {
                    var selectionInfoWindow = document.getElementById(e.diagram.floorplanUI.state.windows.selectionInfoWindow.id);
                    if (selectionInfoWindow.style.visibility !== 'visible')
                        e.diagram.floorplanUI.hideShow('selectionInfoWindow');
                }
            }
        }, new go.Binding("visible", "", function (v, obj) {
            if (obj.part.diagram !== null) {
                return obj.part.diagram.selection.count > 0;
            }
            return false;
        }).ofObject()), $("ContextMenuButton", $(go.TextBlock, "Flip Dimension Side"), {
            click: function (e, obj) {
                var floorplan = obj.part.diagram;
                if (floorplan !== null) {
                    floorplan.startTransaction("flip dimension link side");
                    var walls = [];
                    floorplan.selection.iterator.each(function (part) {
                        if (part.category === "WallGroup")
                            walls.push(part);
                    });
                    for (var i = 0; i < walls.length; i++) {
                        var wall = walls[i];
                        var sPt = wall.data.startpoint.copy();
                        var ePt = wall.data.endpoint.copy();
                        floorplan.model.setDataProperty(wall.data, "startpoint", ePt);
                        floorplan.model.setDataProperty(wall.data, "endpoint", sPt);
                        floorplan.updateWall(wall);
                    }
                    floorplan.commitTransaction("flip dimension link side");
                }
            }
        }, new go.Binding("visible", "", function (v, obj) {
            if (obj.part.diagram !== null) {
                var sel = obj.part.diagram.selection;
                if (sel.count === 0)
                    return false;
                var flag = false;
                sel.iterator.each(function (part) {
                    if (part.category === "WallGroup")
                        flag = true;
                });
                return flag;
            }
            return false;
        }).ofObject()));
    }
    function makeDefaultGroup() {
        var $ = go.GraphObject.make;
        return $(go.Group, "Vertical", {
            contextMenu: makeContextMenu(),
            doubleClick: function (e) {
                if (e.diagram.floorplanUI)
                    e.diagram.floorplanUI.hideShow("selectionInfoWindow");
            },
            toolTip: makeGroupToolTip()
        }, new go.Binding("location", "loc"), $(go.Panel, "Auto", $(go.Shape, "RoundedRectangle", { fill: "rgba(128,128,128,0.15)", stroke: 'rgba(128, 128, 128, .05)', name: 'SHAPE', strokeCap: 'square' }, new go.Binding("fill", "isSelected", function (s, obj) {
            return s ? "rgba(128, 128, 128, .15)" : "rgba(128, 128, 128, 0.10)";
        }).ofObject()), $(go.Placeholder, { padding: 5 })));
    }
    function makeArc(node) {
        var ang = node.data.angle;
        var sweep = node.data.sweep;
        var rad = Math.min(30, node.data.maxRadius);
        if (typeof sweep === "number" && sweep > 0) {
            var start = new go.Point(rad, 0).rotate(ang);
            return new go.Geometry()
                .add(new go.PathFigure(start.x + rad, start.y + rad)
                .add(new go.PathSegment(go.PathSegment.Arc, ang, sweep, rad, rad, rad, rad)))
                .add(new go.PathFigure(0, 0))
                .add(new go.PathFigure(2 * rad, 2 * rad));
        }
        else {
            return new go.Geometry()
                .add(new go.PathFigure(0, 0))
                .add(new go.PathFigure(2 * rad, 2 * rad));
        }
    }
    function makePointNode() {
        var $ = go.GraphObject.make;
        return $(go.Node, "Position", new go.Binding("location", "loc", go.Point.parse).makeTwoWay(go.Point.stringify));
    }
    function makeAngleNode() {
        var $ = go.GraphObject.make;
        return $(go.Node, "Spot", { locationSpot: go.Spot.Center, locationObjectName: "SHAPE", selectionAdorned: false }, new go.Binding("location", "loc", go.Point.parse).makeTwoWay(go.Point.stringify), $(go.Shape, "Circle", { name: "SHAPE", height: 0, width: 0 }), $(go.Shape, { strokeWidth: 1.5, fill: null }, new go.Binding("geometry", "", makeArc).ofObject(), new go.Binding("stroke", "sweep", function (sweep) {
            return (sweep % 45 < 1 || sweep % 45 > 44) ? "dodgerblue" : "lightblue";
        })), $(go.Panel, "Auto", { name: "ARCLABEL" }, new go.Binding("alignment", "sweep", function (sweep, panel) {
            var rad = Math.min(30, panel.part.data.maxRadius);
            var angle = panel.part.data.angle;
            var cntr = new go.Point(rad, 0).rotate(angle + sweep / 2);
            return new go.Spot(0.5, 0.5, cntr.x, cntr.y);
        }), $(go.Shape, { fill: "white" }, new go.Binding("stroke", "sweep", function (sweep) {
            return (sweep % 45 < 1 || sweep % 45 > 44) ? "dodgerblue" : "lightblue";
        })), $(go.TextBlock, { font: "7pt sans-serif", margin: 2 }, new go.Binding("text", "sweep", function (sweep) {
            return sweep.toFixed(2) + String.fromCharCode(176);
        }))));
    }
    function makeDimensionLink() {
        var $ = go.GraphObject.make;
        return $(go.Link, $(go.Shape, { stroke: "gray", strokeWidth: 2, name: 'SHAPE' }), $(go.Shape, { toArrow: "OpenTriangle", stroke: "gray", strokeWidth: 2 }), $(go.Shape, { fromArrow: "BackwardOpenTriangle", stroke: "gray", strokeWidth: 2 }), $(go.TextBlock, { text: 'sometext', segmentOffset: new go.Point(0, -10), font: "13px sans-serif" }, new go.Binding("text", "", function (link) {
            var floorplan = link.diagram;
            if (floorplan) {
                var fromPtNode = null;
                var toPtNode = null;
                floorplan.pointNodes.iterator.each(function (node) {
                    if (node.data.key === link.data.from)
                        fromPtNode = node;
                    if (node.data.key === link.data.to)
                        toPtNode = node;
                });
                if (fromPtNode !== null) {
                    var fromPt = fromPtNode.location;
                    var toPt = toPtNode.location;
                    return floorplan.convertPixelsToUnits(Math.sqrt(fromPt.distanceSquaredPoint(toPt))).toFixed(2) + floorplan.model.modelData.unitsAbbreviation;
                }
                return null;
            }
            return null;
        }).ofObject(), new go.Binding("angle", "angle", function (angle, link) {
            if (angle > 90 && angle < 270)
                return (angle + 180) % 360;
            return angle;
        }), new go.Binding("segmentOffset", "angle", function (angle, textblock) {
            var floorplan = textblock.part.diagram;
            if (floorplan) {
                var wall = floorplan.findPartForKey(textblock.part.data.wall);
                if (wall.rotateObject.angle > 135 && wall.rotateObject.angle < 315)
                    return new go.Point(0, 10);
                return new go.Point(0, -10);
            }
            return new go.Point(0, 0);
        }).ofObject(), new go.Binding("font", "", function (link) {
            var floorplan = link.diagram;
            var fromPtNode = null;
            var toPtNode = null;
            floorplan.pointNodes.iterator.each(function (node) {
                if (node.data.key === link.data.from)
                    fromPtNode = node;
                if (node.data.key === link.data.to)
                    toPtNode = node;
            });
            if (fromPtNode !== null) {
                var fromPt = fromPtNode.location;
                var toPt = toPtNode.location;
                var distance = Math.sqrt(fromPt.distanceSquaredPoint(toPt));
                if (distance > 40)
                    return "13px sans-serif";
                if (distance <= 40 && distance >= 20)
                    return "11px sans-serif";
                else
                    return "9px sans-serif";
            }
            return "13px sans-serif";
        }).ofObject()));
    }
    function makeNodeToolTip() {
        var $ = go.GraphObject.make;
        return $(go.Adornment, "Auto", $(go.Shape, { fill: "#FFFFCC" }), $(go.TextBlock, { margin: 4 }, new go.Binding("text", "", function (text, obj) {
            var data = obj.part.adornedObject.data;
            var name = (obj.part.adornedObject.category === "MultiPurposeNode") ? data.text : data.caption;
            return "Name: " + name + "\nNotes: " + data.notes;
        }).ofObject()));
    }
    function makeFurnitureResizeAdornmentTemplate() {
        var $ = go.GraphObject.make;
        function makeHandle(alignment, cursor) {
            return $(go.Shape, { alignment: alignment, cursor: cursor, figure: "Rectangle", desiredSize: new go.Size(7, 7), fill: "#ffffff", stroke: "#808080" }, new go.Binding("fill", "color"), new go.Binding("stroke", "stroke"));
        }
        return $(go.Adornment, "Spot", $(go.Placeholder), makeHandle(go.Spot.Top, "n-resize"), makeHandle(go.Spot.TopRight, "n-resize"), makeHandle(go.Spot.BottomRight, "se-resize"), makeHandle(go.Spot.Right, "e-resize"), makeHandle(go.Spot.Bottom, "s-resize"), makeHandle(go.Spot.BottomLeft, "sw-resize"), makeHandle(go.Spot.Left, "w-resize"), makeHandle(go.Spot.TopLeft, "nw-resize"));
    }
    function makeFurnitureRotateAdornmentTemplate() {
        var $ = go.GraphObject.make;
        return $(go.Adornment, $(go.Shape, "Circle", { cursor: "pointer", desiredSize: new go.Size(7, 7), fill: "#ffffff", stroke: "#808080" }, new go.Binding("fill", "", function (obj) { return (obj.adornedPart === null) ? "#ffffff" : obj.adornedPart.data.color; }).ofObject(), new go.Binding("stroke", "", function (obj) { return (obj.adornedPart === null) ? "#000000" : obj.adornedPart.data.stroke; }).ofObject()));
    }
    function invertColor(hex) {
        if (hex.indexOf('#') === 0) {
            hex = hex.slice(1);
        }
        if (hex.length === 3) {
            hex = hex[0] + hex[0] + hex[1] + hex[1] + hex[2] + hex[2];
        }
        if (hex.length !== 6) {
            throw new Error('Invalid HEX color.');
        }
        var r = (255 - parseInt(hex.slice(0, 2), 16)).toString(16), g = (255 - parseInt(hex.slice(2, 4), 16)).toString(16), b = (255 - parseInt(hex.slice(4, 6), 16)).toString(16);
        return '#' + padZero(r) + padZero(g) + padZero(b);
    }
    function padZero(str) {
        var len = 2;
        var zeros = new Array(len).join('0');
        return (zeros + str).slice(-len);
    }
    function makeDefaultNode() {
        var $ = go.GraphObject.make;
        return $(go.Node, "Spot", {
            resizable: true,
            rotatable: true,
            toolTip: makeNodeToolTip(),
            resizeAdornmentTemplate: makeFurnitureResizeAdornmentTemplate(),
            rotateAdornmentTemplate: makeFurnitureRotateAdornmentTemplate(),
            contextMenu: makeContextMenu(),
            locationObjectName: "SHAPE",
            resizeObjectName: "SHAPE",
            rotateObjectName: "SHAPE",
            minSize: new go.Size(5, 5),
            locationSpot: go.Spot.Center,
            selectionAdorned: false,
            doubleClick: function (e) {
                if (e.diagram.floorplanUI)
                    e.diagram.floorplanUI.hideShow("selectionInfoWindow");
            }
        }, new go.Binding("location", "loc", go.Point.parse).makeTwoWay(go.Point.stringify), new go.Binding("layerName", "isSelected", function (s) {
            return s ? "Foreground" : "";
        }).ofObject(), $(go.Shape, {
            name: "SHAPE", stroke: "#000000",
            fill: "rgba(128, 128, 128, 0.5)"
        }, new go.Binding("geometryString", "geo"), new go.Binding("width").makeTwoWay(), new go.Binding("height").makeTwoWay(), new go.Binding("angle").makeTwoWay(), new go.Binding("fill", "color")), new go.Binding("stroke", "isSelected", function (s, obj) {
            return s ? go.Brush.lightenBy(obj.stroke, .5) : invertColor(obj.part.data.color);
        }).ofObject());
    }
    function makeMultiPurposeNode() {
        var $ = go.GraphObject.make;
        return $(go.Node, "Spot", {
            contextMenu: makeContextMenu(),
            toolTip: makeNodeToolTip(),
            locationSpot: go.Spot.Center,
            resizeAdornmentTemplate: makeFurnitureResizeAdornmentTemplate(),
            rotateAdornmentTemplate: makeFurnitureRotateAdornmentTemplate(),
            locationObjectName: "SHAPE",
            resizable: true,
            rotatable: true,
            resizeObjectName: "SHAPE",
            rotateObjectName: "SHAPE",
            minSize: new go.Size(5, 5),
            selectionAdorned: false,
            doubleClick: function (e) {
                if (e.diagram.floorplanUI)
                    e.diagram.floorplanUI.hideShow("selectionInfoWindow");
            }
        }, new go.Binding("location", "loc", go.Point.parse).makeTwoWay(go.Point.stringify), new go.Binding("layerName", "isSelected", function (s) { return s ? "Foreground" : ""; }).ofObject(), $(go.Shape, { strokeWidth: 1, name: "SHAPE", fill: "rgba(128, 128, 128, 0.5)", }, new go.Binding("angle").makeTwoWay(), new go.Binding("width").makeTwoWay(), new go.Binding("height").makeTwoWay(), new go.Binding("fill", "color"), new go.Binding("stroke", "isSelected", function (s, obj) {
            return s ? go.Brush.lightenBy(obj.stroke, .5) : invertColor(obj.part.data.color);
        }).ofObject()), $(go.TextBlock, {
            margin: 5,
            wrap: go.TextBlock.WrapFit,
            textAlign: "center",
            editable: true,
            isMultiline: false,
            stroke: '#454545',
            font: "10pt sans-serif"
        }, new go.Binding("text").makeTwoWay(), new go.Binding("angle", "angle").makeTwoWay(), new go.Binding("font", "height", function (height) {
            if (height > 25)
                return "10pt sans-serif";
            if (height < 25 && height > 15)
                return "8pt sans-serif";
            else
                return "6pt sans-serif";
        }), new go.Binding("stroke", "color", function (color) { return invertColor(color); })));
    }
    var snapWalls = function (part, pt, gridPt) {
        var floorplan = part.diagram;
        floorplan.updateWallDimensions();
        floorplan.updateWallAngles();
        floorplan.updateWall(part);
        var grid = part.diagram.grid;
        var sPt = part.data.startpoint.copy();
        var ePt = part.data.endpoint.copy();
        var dx = pt.x - part.location.x;
        var dy = pt.y - part.location.y;
        var newSpt = sPt.offset(dx, dy);
        var newEpt = ePt.offset(dx, dy);
        if (floorplan.toolManager.draggingTool.isGridSnapEnabled) {
            newSpt = newSpt.snapToGridPoint(grid.gridOrigin, grid.gridCellSize);
            newEpt = newEpt.snapToGridPoint(grid.gridOrigin, grid.gridCellSize);
        }
        floorplan.model.setDataProperty(part.data, "startpoint", newSpt);
        floorplan.model.setDataProperty(part.data, "endpoint", newEpt);
        return new go.Point((newSpt.x + newEpt.x) / 2, (newSpt.y + newEpt.y) / 2);
    };
    function findClosestLocOnWall(wall, part) {
        var orderedConstrainingPts = [];
        var startpoint = wall.data.startpoint.copy();
        var endpoint = wall.data.endpoint.copy();
        var firstWallPt = ((startpoint.x + startpoint.y) <= (endpoint.x + endpoint.y)) ? startpoint : endpoint;
        var lastWallPt = ((startpoint.x + startpoint.y) > (endpoint.x + endpoint.y)) ? startpoint : endpoint;
        var wallPartEndpoints = [];
        wall.memberParts.iterator.each(function (wallPart) {
            var endpoints = getWallPartEndpoints(wallPart);
            wallPartEndpoints.push(endpoints[0]);
            wallPartEndpoints.push(endpoints[1]);
        });
        wallPartEndpoints.sort(function (a, b) {
            if ((a.x + a.y) > (b.x + b.y))
                return 1;
            if ((a.x + a.y) < (b.x + b.y))
                return -1;
            else
                return 0;
        });
        orderedConstrainingPts.push(firstWallPt);
        orderedConstrainingPts = orderedConstrainingPts.concat(wallPartEndpoints);
        orderedConstrainingPts.push(lastWallPt);
        var possibleStretches = [];
        for (var i = 0; i < orderedConstrainingPts.length; i += 2) {
            var point1 = orderedConstrainingPts[i];
            var point2 = orderedConstrainingPts[i + 1];
            var distanceBetween = Math.sqrt(point1.distanceSquaredPoint(point2));
            if (distanceBetween >= part.data.length)
                possibleStretches.push({ pt1: point1, pt2: point2 });
        }
        var closestDist = Number.MAX_VALUE;
        var closestStretch = null;
        for (var i = 0; i < possibleStretches.length; i++) {
            var testStretch = possibleStretches[i];
            var testPoint1 = testStretch.pt1;
            var testPoint2 = testStretch.pt2;
            var testDistance1 = Math.sqrt(testPoint1.distanceSquaredPoint(part.location));
            var testDistance2 = Math.sqrt(testPoint2.distanceSquaredPoint(part.location));
            if (testDistance1 < closestDist) {
                closestDist = testDistance1;
                closestStretch = testStretch;
            }
            if (testDistance2 < closestDist) {
                closestDist = testDistance2;
                closestStretch = testStretch;
            }
        }
        if (closestStretch === null)
            return null;
        var closestStretchLength = Math.sqrt(closestStretch.pt1.distanceSquaredPoint(closestStretch.pt2));
        var offset = part.data.length / 2;
        var pt1 = new go.Point(closestStretch.pt1.x + ((offset / closestStretchLength) * (closestStretch.pt2.x - closestStretch.pt1.x)), closestStretch.pt1.y + ((offset / closestStretchLength) * (closestStretch.pt2.y - closestStretch.pt1.y)));
        var pt2 = new go.Point(closestStretch.pt2.x + ((offset / closestStretchLength) * (closestStretch.pt1.x - closestStretch.pt2.x)), closestStretch.pt2.y + ((offset / closestStretchLength) * (closestStretch.pt1.y - closestStretch.pt2.y)));
        var newLoc = part.location.copy().projectOntoLineSegmentPoint(pt1, pt2);
        return newLoc;
    }
    var addWallPart = function (e, wall) {
        var floorplan = e.diagram;
        var wallPart = floorplan.selection.first();
        if ((wallPart && (wallPart.category === "WindowNode" || wallPart.category === "DoorNode") && wallPart.containingGroup === null)) {
            var newLoc = findClosestLocOnWall(wall, wallPart);
            if (newLoc !== null) {
                wall.findObject("SHAPE").stroke = "black";
                floorplan.model.setDataProperty(wallPart.data, "group", wall.data.key);
                wallPart.location = newLoc.projectOntoLineSegmentPoint(wall.data.startpoint, wall.data.endpoint);
                wallPart.angle = wall.rotateObject.angle;
                if (wallPart.category === "WindowNode")
                    floorplan.model.setDataProperty(wallPart.data, "height", wall.data.thickness);
                if (wallPart.category === "DoorNode")
                    floorplan.model.setDataProperty(wallPart.data, "doorOpeningHeight", wall.data.thickness);
            }
            else {
                floorplan.remove(wallPart);
                alert("There's not enough room on the wall!");
                return;
            }
        }
        if (floorplan.floorplanUI)
            floorplan.floorplanUI.setSelectionInfo(floorplan.selection.first(), floorplan);
        floorplan.updateWallDimensions();
    };
    var wallPartDragOver = function (e, wall) {
        var floorplan = e.diagram;
        var parts = floorplan.toolManager.draggingTool.draggingParts;
        parts.iterator.each(function (part) {
            if ((part.category === "WindowNode" || part.category === "DoorNode") && part.containingGroup === null) {
                wall.findObject("SHAPE").stroke = "lightblue";
                part.angle = wall.rotateObject.angle;
            }
        });
    };
    var wallPartDragAway = function (e, wall) {
        var floorplan = e.diagram;
        wall.findObject("SHAPE").stroke = "black";
        var parts = floorplan.toolManager.draggingTool.draggingParts;
        parts.iterator.each(function (part) {
            if ((part.category === "WindowNode" || part.category === "DoorNode") && part.containingGroup === null)
                part.angle = 0;
        });
    };
    function makeWallGroup() {
        var $ = go.GraphObject.make;
        return $(go.Group, "Spot", {
            contextMenu: makeContextMenu(),
            toolTip: makeGroupToolTip(),
            selectionObjectName: "SHAPE",
            rotateObjectName: "SHAPE",
            locationSpot: go.Spot.Center,
            reshapable: true,
            minSize: new go.Size(1, 1),
            dragComputation: snapWalls,
            selectionAdorned: false,
            mouseDrop: addWallPart,
            mouseDragEnter: wallPartDragOver,
            mouseDragLeave: wallPartDragAway,
            doubleClick: function (e) { if (e.diagram.floorplanUI)
                e.diagram.floorplanUI.hideShow("selectionInfoWindow"); }
        }, $(go.Shape, {
            name: "SHAPE",
            fill: "black",
        }, new go.Binding("strokeWidth", "thickness"), new go.Binding("stroke", "isSelected", function (s, obj) {
            if (obj.part.containingGroup != null) {
                var group = obj.part.containingGroup;
                if (s) {
                    group.data.isSelected = true;
                }
            }
            return s ? "dodgerblue" : "black";
        }).ofObject()));
    }
    function getWallPartEndpoints(wallPart) {
        var loc = wallPart.location;
        var partLength = wallPart.data.length;
        var angle = 0;
        if (wallPart.containingGroup !== null)
            angle = wallPart.containingGroup.rotateObject.angle;
        else
            angle = 180;
        var point1 = new go.Point((loc.x + (partLength / 2)), loc.y);
        var point2 = new go.Point((loc.x - (partLength / 2)), loc.y);
        point1.offset(-loc.x, -loc.y).rotate(angle).offset(loc.x, loc.y);
        point2.offset(-loc.x, -loc.y).rotate(angle).offset(loc.x, loc.y);
        var arr = [];
        arr.push(point1);
        arr.push(point2);
        return arr;
    }
    function getWallPartStretch(part) {
        var wall = part.containingGroup;
        var startpoint = wall.data.startpoint.copy();
        var endpoint = wall.data.endpoint.copy();
        var leftOrAbove = new go.Set();
        var rightOrBelow = new go.Set();
        wall.memberParts.iterator.each(function (wallPart) {
            if (wallPart.data.key !== part.data.key) {
                var endpoints = getWallPartEndpoints(wallPart);
                for (var i = 0; i < endpoints.length; i++) {
                    if (endpoints[i].x < part.location.x || (endpoints[i].y > part.location.y && endpoints[i].x === part.location.x))
                        leftOrAbove.add(endpoints[i]);
                    else
                        rightOrBelow.add(endpoints[i]);
                }
            }
        });
        if (parseFloat(startpoint.x.toFixed(2)) < parseFloat(part.location.x.toFixed(2)) || (startpoint.y > part.location.y && parseFloat(startpoint.x.toFixed(2)) === parseFloat(part.location.x.toFixed(2))))
            leftOrAbove.add(startpoint);
        else
            rightOrBelow.add(startpoint);
        if (parseFloat(endpoint.x.toFixed(2)) < parseFloat(part.location.x.toFixed(2)) || (endpoint.y > part.location.y && parseFloat(endpoint.x.toFixed(2)) === parseFloat(part.location.x.toFixed(2))))
            leftOrAbove.add(endpoint);
        else
            rightOrBelow.add(endpoint);
        var leftOrAbovePt;
        var closestDistLeftOrAbove = Number.MAX_VALUE;
        leftOrAbove.iterator.each(function (point) {
            var pt = point;
            var distance = Math.sqrt(pt.distanceSquaredPoint(part.location));
            if (distance < closestDistLeftOrAbove) {
                closestDistLeftOrAbove = distance;
                leftOrAbovePt = pt;
            }
        });
        var rightOrBelowPt;
        var closestDistRightOrBelow = Number.MAX_VALUE;
        rightOrBelow.iterator.each(function (point) {
            var pt = point;
            var distance = Math.sqrt(pt.distanceSquaredPoint(part.location));
            if (distance < closestDistRightOrBelow) {
                closestDistRightOrBelow = distance;
                rightOrBelowPt = pt;
            }
        });
        var stretch = { point1: leftOrAbovePt, point2: rightOrBelowPt };
        return stretch;
    }
    var dragWallParts = function (part, pt, gridPt) {
        if (part.containingGroup !== null && part.containingGroup.category === 'WallGroup') {
            var floorplan = part.diagram;
            var wall = part.containingGroup;
            var wStart = wall.data.startpoint;
            var wEnd = wall.data.endpoint;
            var dist1 = Math.sqrt(wStart.distanceSquaredPoint(part.location));
            var dist2 = Math.sqrt(part.location.distanceSquaredPoint(wEnd));
            var totalDist = Math.sqrt(wStart.distanceSquaredPoint(wEnd));
            if (dist1 + dist2 !== totalDist)
                part.location = part.location.copy().projectOntoLineSegmentPoint(wStart, wEnd);
            var stretch = getWallPartStretch(part);
            var leftOrAbovePt = stretch.point1;
            var rightOrBelowPt = stretch.point2;
            var totalLength = Math.sqrt(leftOrAbovePt.distanceSquaredPoint(rightOrBelowPt));
            var distance = (part.data.length / 2);
            var point1 = new go.Point(leftOrAbovePt.x + ((distance / totalLength) * (rightOrBelowPt.x - leftOrAbovePt.x)), leftOrAbovePt.y + ((distance / totalLength) * (rightOrBelowPt.y - leftOrAbovePt.y)));
            var point2 = new go.Point(rightOrBelowPt.x + ((distance / totalLength) * (leftOrAbovePt.x - rightOrBelowPt.x)), rightOrBelowPt.y + ((distance / totalLength) * (leftOrAbovePt.y - rightOrBelowPt.y)));
            var distFromWall = Math.abs(((wEnd.y - wStart.y) * pt.x) - ((wEnd.x - wStart.x) * pt.y) + (wEnd.x * wStart.y) - (wEnd.y * wStart.x)) /
                Math.sqrt(Math.pow((wEnd.y - wStart.y), 2) + Math.pow((wEnd.x - wStart.x), 2));
            var tolerance = (20 * wall.data.thickness < 100) ? (20 * wall.data.thickness) : 100;
            if (distFromWall > tolerance) {
                part.containingGroup = null;
                delete part.data.group;
                part.angle = 0;
                floorplan.pointNodes.iterator.each(function (node) { floorplan.remove(node); });
                floorplan.dimensionLinks.iterator.each(function (link) { floorplan.remove(link); });
                floorplan.pointNodes.clear();
                floorplan.dimensionLinks.clear();
                floorplan.updateWallDimensions();
                if (floorplan.floorplanUI)
                    floorplan.floorplanUI.setSelectionInfo(part);
            }
            pt = pt.copy().projectOntoLineSegmentPoint(point1, point2);
            floorplan.skipsUndoManager = true;
            floorplan.startTransaction("set loc");
            floorplan.model.setDataProperty(part.data, "loc", go.Point.stringify(pt));
            floorplan.commitTransaction("set loc");
            floorplan.skipsUndoManager = false;
            floorplan.updateWallDimensions();
        }
        return pt;
    };
    function makeWallPartResizeAdornment() {
        var $ = go.GraphObject.make;
        return $(go.Adornment, "Spot", { name: "WallPartResizeAdornment" }, $(go.Placeholder), $(go.Shape, { alignment: go.Spot.Left, cursor: "w-resize", figure: "Diamond", desiredSize: new go.Size(7, 7), fill: "#ffffff", stroke: "#808080" }), $(go.Shape, { alignment: go.Spot.Right, cursor: "e-resize", figure: "Diamond", desiredSize: new go.Size(7, 7), fill: "#ffffff", stroke: "#808080" }));
    }
    function makeDoorSelectionAdornment() {
        var $ = go.GraphObject.make;
        return $(go.Adornment, "Vertical", { name: "DoorSelectionAdornment" }, $(go.Panel, "Auto", $(go.Shape, { fill: null, stroke: null }), $(go.Placeholder)), $(go.Panel, "Horizontal", { defaultStretch: go.GraphObject.Vertical }, $("Button", $(go.Picture, { source: "icons/flipDoorOpeningLeft.png", column: 0, desiredSize: new go.Size(12, 12) }, new go.Binding("source", "", function (obj) {
            if (obj.adornedPart === null)
                return "icons/flipDoorOpeningRight.png";
            else if (obj.adornedPart.data.swing === "left")
                return "icons/flipDoorOpeningRight.png";
            else
                return "icons/flipDoorOpeningLeft.png";
        }).ofObject()), {
            click: function (e, obj) {
                var floorplan = obj.part.diagram;
                floorplan.startTransaction("flip door");
                var door = obj.part.adornedPart;
                if (door.data.swing === "left")
                    floorplan.model.setDataProperty(door.data, "swing", "right");
                else
                    floorplan.model.setDataProperty(door.data, "swing", "left");
                floorplan.commitTransaction("flip door");
            },
            toolTip: $(go.Adornment, "Auto", $(go.Shape, { fill: "#FFFFCC" }), $(go.TextBlock, { margin: 4, text: "Flip Door Opening" }))
        }, new go.Binding("visible", "", function (obj) { return (obj.adornedPart === null) ? false : (obj.adornedPart.containingGroup !== null); }).ofObject()), $("Button", $(go.Picture, { source: "icons/flipDoorSide.png", column: 0, desiredSize: new go.Size(12, 12) }), {
            click: function (e, obj) {
                var floorplan = obj.part.diagram;
                floorplan.startTransaction("rotate door");
                var door = obj.part.adornedPart;
                door.angle = (door.angle + 180) % 360;
                floorplan.commitTransaction("rotate door");
            },
            toolTip: $(go.Adornment, "Auto", $(go.Shape, { fill: "#FFFFCC" }), $(go.TextBlock, { margin: 4, text: "Flip Door Side" }))
        }), new go.Binding("visible", "", function (obj) { return (obj.adornedPart === null) ? false : (obj.adornedPart.containingGroup !== null); }).ofObject()));
    }
    function makeWindowNode() {
        var $ = go.GraphObject.make;
        return $(go.Node, "Spot", {
            contextMenu: makeContextMenu(),
            selectionObjectName: "SHAPE",
            selectionAdorned: false,
            locationSpot: go.Spot.Center,
            toolTip: makeNodeToolTip(),
            minSize: new go.Size(5, 5),
            resizable: true,
            resizeAdornmentTemplate: makeWallPartResizeAdornment(),
            resizeObjectName: "SHAPE",
            rotatable: false,
            doubleClick: function (e) { if (e.diagram.floorplanUI)
                e.diagram.floorplanUI.hideShow("selectionInfoWindow"); },
            dragComputation: dragWallParts,
            layerName: 'Foreground'
        }, new go.Binding("location", "loc", go.Point.parse).makeTwoWay(go.Point.stringify), new go.Binding("angle").makeTwoWay(), $(go.Shape, { name: "SHAPE", fill: "white", strokeWidth: 0 }, new go.Binding("width", "length").makeTwoWay(), new go.Binding("height").makeTwoWay(), new go.Binding("stroke", "isSelected", function (s, obj) { return s ? "dodgerblue" : "black"; }).ofObject(), new go.Binding("fill", "isSelected", function (s, obj) { return s ? "lightgray" : "white"; }).ofObject()), $(go.Shape, { name: "LINESHAPE", fill: "darkgray", strokeWidth: 0, height: 10 }, new go.Binding("width", "length", function (width, obj) { return width - 10; }), new go.Binding("height", "height", function (height, obj) { return (height / 5); }), new go.Binding("stroke", "isSelected", function (s, obj) { return s ? "dodgerblue" : "black"; }).ofObject()));
    }
    function makeDoorNode() {
        var $ = go.GraphObject.make;
        return $(go.Node, "Spot", {
            contextMenu: makeContextMenu(),
            selectionObjectName: "SHAPE",
            selectionAdornmentTemplate: makeDoorSelectionAdornment(),
            locationSpot: go.Spot.BottomCenter,
            resizable: true,
            resizeObjectName: "OPENING_SHAPE",
            toolTip: makeNodeToolTip(),
            minSize: new go.Size(10, 10),
            doubleClick: function (e) { if (e.diagram.floorplanUI)
                e.diagram.floorplanUI.hideShow("selectionInfoWindow"); },
            dragComputation: dragWallParts,
            resizeAdornmentTemplate: makeWallPartResizeAdornment(),
            layerName: 'Foreground'
        }, new go.Binding("location", "loc", go.Point.parse).makeTwoWay(go.Point.stringify), new go.Binding("angle").makeTwoWay(), new go.Binding("locationSpot", "doorOpeningHeight", function (doh, obj) { return new go.Spot(0.5, 1, 0, -(doh / 2)); }), $(go.Shape, { name: "SHAPE" }, new go.Binding("width", "length"), new go.Binding("height", "length").makeTwoWay(), new go.Binding("stroke", "isSelected", function (s, obj) { return s ? "dodgerblue" : "black"; }).ofObject(), new go.Binding("fill", "color"), new go.Binding("geometryString", "swing", function (swing) {
            if (swing === "left")
                return "F1 M0,0 v-150 a150,150 0 0,1 150,150 ";
            else
                return "F1 M275,175 v-150 a150,150 0 0,0 -150,150 ";
        })), $(go.Shape, {
            name: "OPENING_SHAPE", fill: "white",
            strokeWidth: 0, height: 5, width: 40,
            alignment: go.Spot.BottomCenter, alignmentFocus: go.Spot.Center
        }, new go.Binding("height", "doorOpeningHeight").makeTwoWay(), new go.Binding("stroke", "isSelected", function (s, obj) { return s ? "dodgerblue" : "black"; }).ofObject(), new go.Binding("fill", "isSelected", function (s, obj) { return s ? "lightgray" : "white"; }).ofObject(), new go.Binding("width", "length").makeTwoWay()));
    }
    function makePaletteWallNode() {
        var $ = go.GraphObject.make;
        return $(go.Node, "Spot", { selectionAdorned: false }, $(go.Shape, { name: "SHAPE", fill: "black", strokeWidth: 0, height: 10, figure: "Rectangle" }, new go.Binding("width", "length").makeTwoWay(), new go.Binding("height").makeTwoWay(), new go.Binding("fill", "isSelected", function (s, obj) { return s ? "dodgerblue" : "black"; }).ofObject(), new go.Binding("stroke", "isSelected", function (s, obj) { return s ? "dodgerblue" : "black"; }).ofObject()));
    }
    return Floorplan;
});
